@@ -2,41 +2,53 @@ from collections import defaultdict
 import pandas as pd
 import numpy as np
 from time import time
+import logging
 
 class Recommender(object):
     """
     Cold Start Recommender
     """
-    def __init__(self, mongo_host=None, mongo_db_name=None, mongo_replica_set=None, default_rating=3, min_rating=1, max_rating=5):
+    def __init__(self, mongo_host=None, mongo_db_name=None, mongo_replica_set=None,
+                 default_rating=3, 
+                 log_level=logging.ERROR):
         if mongo_host is not None:
-            from pymongo import MongoClient
             assert (mongo_db_name != None)
             if mongo_replica_set is not None:
-                from pymongo import MongoReplicaSetClient
                 mongo_client = MongoReplicaSetClient(mongo_host, replicaSet=mongo_replica_set)
                 self.db = mongo_client[mongo_db_name]
+                # reading --for producing recommendations-- could be even out of sync.
+                # this can be added if most replicas are in-memory
+                # self.db.read_preference = ReadPreference.SECONDARY_PREFERRED
             else:
                 mongo_client = MongoClient(mongo_host)
                 self.db = mongo_client[mongo_db_name]
+            # If these tables do not exist, it might create problems
+            if not self.db['user_ratings'].find_one():
+                self.db['user_ratings'].insert({})
+            if not self.db['item_ratings'].find_one():
+                self.db['item_ratings'].insert({})
         else:
             self.db = None
-        self.info_used = set() # Info used in addition to item_id. It is filled each time an item is inserted with item_info
-        self.default_rating = default_rating
-        self.min_rating = min_rating
-        self.max_rating = max_rating
+        self.info_used = set() # Info used in addition to item_id. Only for in-memory testing, otherwise there is utils collection in the MongoDB
+        self.default_rating = default_rating  # Rating inserted by default
         self._items_cooccurence = pd.DataFrame  # cooccurrence of items
         self._categories_cooccurence = {} # cooccurrence of categories
-        self.cooccurence_updated = 0.0
-        self.item_ratings = defaultdict(dict)  # matrix of ratings for a item
-        self.user_ratings = defaultdict(dict)  # matrix of ratings for a user
-        self.items = defaultdict(dict)  # matrix of item's information {item_id: {"cat1": "my_cat"....}
+        self.cooccurence_updated = 0.0  # Time of update
+        self.item_ratings = defaultdict(dict)  # matrix of ratings for a item (inmemory testing)
+        self.user_ratings = defaultdict(dict)  # matrix of ratings for a user (inmemory testing)
+        self.items = defaultdict(dict)  # matrix of item's information {item_id: {"Author": "AA. VV."....}
         # categories --same as above, but separated as they are not always available
-        self.tot_categories_user_ratings = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
-        self.tot_categories_item_ratings = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
-        self.n_categories_user_ratings = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
-        self.n_categories_item_ratings = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))
+        self.tot_categories_user_ratings = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))  # sum of all ratings  (inmemory testing)
+        self.tot_categories_item_ratings = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))  # ditto
+        self.n_categories_user_ratings = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))  # number of ratings  (inmemory testing)
+        self.n_categories_item_ratings = defaultdict(lambda: defaultdict(lambda: defaultdict(int)))  # ditto
         self.items_by_popularity = []
-        self.items_by_popularity_updated = 0.0
+        self.items_by_popularity_updated = 0.0  # Time of update
+        # Loggin stuff
+
+        self.logger = logging.getLogger('CSREC')
+        self.logger.setLevel(log_level)
+        self.logger.debug("============ Creating a Recommender Instance ================")
 
     def _coll_name(self, k, typ):
         """
@@ -51,32 +63,69 @@ class Recommender(object):
         :return:
         """
         df_tot_cat_item = {}
-        # TO BE DELETED df_n_cat_item = {}
         if not self.db:
             # Items' vectors
             df_item = pd.DataFrame(self.item_ratings).fillna(0).astype(int)
             # Categories' vectors
             if len(self.info_used) > 0:
-                for i in self.info_used:
+                info_used = self.info_used
+                for i in info_used:
                     df_tot_cat_item[i] = pd.DataFrame(self.tot_categories_item_ratings[i]).fillna(0).astype(int)
         else:  # read if from Mongodb
             # here we *must* use user_ratings, so indexes are the users, columns the items...
             df_item = pd.DataFrame.from_records(list(self.db['user_ratings'].find())).set_index('_id').fillna(0).astype(int)
-            if len(self.info_used) > 0:
-                for i in self.info_used:
+            info_used = self.db['utils'].find_one({"_id": 1}, {'info_used': 1, "_id": 0}).get('info_used', [])
+            self.logger.info("[_create_cooccurence] Found info_used in db.utils: %s", info_used)
+            if len(info_used) > 0:
+                for i in info_used:
                     user_coll_name = self._coll_name(i, 'user')
                     if self.db['tot_' + user_coll_name].find_one():
                         df_tot_cat_item[i] = pd.DataFrame.from_records(list(self.db['tot_' + user_coll_name].find())).set_index('_id').fillna(0).astype(int)
         df_item = (df_item / df_item).replace(np.inf, 0)  # normalize to one to build the co-occurence
         self._items_cooccurence = df_item.T.dot(df_item)
-        if len(self.info_used) > 0:
-            for i in self.info_used:
+        if len(info_used) > 0:
+            for i in info_used:
                 if type(df_tot_cat_item.get(i)) == pd.DataFrame:
                     df_tot_cat_item[i] = (df_tot_cat_item[i] / df_tot_cat_item[i]).replace(np.inf, 0)
                     self._categories_cooccurence[i] = df_tot_cat_item[i].T.dot(df_tot_cat_item[i])
         self.cooccurence_updated = time()
 
 
+    def _sync_user_item_ratings(self):
+        """
+        It might happen that the user_ratings and the item_ratings
+        are not aligned. It shouldn't, but with users can be profiled,
+        then reconciled with session_id etc, it happened...
+        THIS SHOULD BE LOGGED AS ERROR!
+        :return:
+        """
+        #Doing that only for the mongodb case..
+        self.logger.warning("[_sync_user_item_ratings] Syncronyzing item_ratings with user_ratings data")
+        if self.db:
+            self.db['item_ratings'].drop()
+            for user_date in self.db['user_ratings'].find():  # don't put {_id: 0!}
+                for item_id, rating in user_date.iteritems():
+                    if item_id != "_id":
+                        self.db['item_ratings'].update(
+                            {"_id": item_id},
+                            {"$set": {user_date["_id"]: rating}},
+                            upsert=True
+                        )
+            for coll in self.db.collection_names():
+                if '_item_' in coll:
+                    user_coll = coll.replace('_item_', '_user_')
+                    self.db[coll].drop()
+                    for user_date in self.db[user_coll].find():  # don't put {_id: 0!}
+                        for info_id, value in user_date.iteritems():
+                            if info_id != "_id":
+                                self.db[coll].update(
+                                    {"_id": info_id},
+                                    {"$set": {user_date["_id"]: value}},
+                                    upsert=True
+                                )
+
+
+
     def insert_item(self, item, _id="_id"):
         """
         Insert the whole document either in self.items or in db.items
@@ -140,6 +189,8 @@ class Recommender(object):
                         {"$set": {key: value}},
                         upsert=True
                     )
+                    self.logger.info("[reconcile_ids] %s : %s for new user_id %s", key, value, id_new)
+                self.logger.info("[reconcile_ids] Removing %s from user_rating", id_old)
                 self.db['user_ratings'].remove({"_id": id_old})
 
             self.db['item_ratings'].update(
@@ -148,12 +199,14 @@ class Recommender(object):
                 multi=True
             )
 
-            if len(self.info_used) > 0:
-                for k in self.info_used:
+            info_used = self.db['utils'].find_one({"_id": 1}, {'info_used': 1, "_id": 0}).get('info_used', [])
+            self.logger.debug("[reconcile_ids] info_used %s", info_used)
+            if len(info_used) > 0:
+                for k in info_used:
                     users_coll_name = self._coll_name(k, 'user')
                     items_coll_name = self._coll_name(k, 'item')
                     # tot and n user ratings....
-                    tot_user_ratings = self.db['tot_' + users_coll_name].get({"_id": id_old})
+                    tot_user_ratings = self.db['tot_' + users_coll_name].find_one({"_id": id_old}, {"_id": 0})
                     if tot_user_ratings:
                         for key, value in tot_user_ratings.items():
                             self.db['tot_' + users_coll_name].update(
@@ -162,7 +215,7 @@ class Recommender(object):
                                 upsert=True
                             )
                         self.db['tot_' + users_coll_name].remove({"_id": id_old})
-                    n_user_ratings = self.db['n_' + users_coll_name].get({"_id": id_old})
+                    n_user_ratings = self.db['n_' + users_coll_name].find_one({"_id": id_old}, {"_id": 0})
                     if n_user_ratings:
                         for key, value in n_user_ratings.items():
                             self.db['n_' + users_coll_name].update(
@@ -186,7 +239,7 @@ class Recommender(object):
         self._create_cooccurence()
 
 
-    def compute_items_by_popularity(self, fast=False):
+    def compute_items_by_popularity(self, max_items=10, fast=False):
         """
         As per name, get self.
         :return: list of popular items, 0=most popular
@@ -196,13 +249,16 @@ class Recommender(object):
 
         if not self.db:
             df_item = pd.DataFrame(self.item_ratings).fillna(0).astype(int).sum()
-
-        else:  # read if from Mongodb
+        else:  # Mongodb
             df_item = pd.DataFrame.from_records(list(self.db['user_ratings'].find())).set_index('_id').sum()
 
         df_item.sort(ascending=False)
-        self.items_by_popularity = list(df_item.index)
-
+        pop_items = list(df_item.index)
+        if len(pop_items) >= max_items:
+            self.items_by_popularity = pop_items
+        else:
+            all_items = set([ d["_id"] for d in self.db['items'].find({}, {"_id": 1})])
+            self.items_by_popularity = pop_items + list( all_items - set(pop_items) )
 
     def get_similar_item(self, item_id, user_id=None, algorithm='simple'):
         """
@@ -301,15 +357,22 @@ class Recommender(object):
                 self.item_ratings[item_id][user_id] = float(rating)
         # MongoDB
         else:
-            # Do categories only if the item is found stored
+            # If the item is not stored, we don't have its categories
+            # Therefore do categories only if the item is found stored
             if self.db['items'].find_one({"_id": item_id}):
                 item = self.db['items'].find_one({"_id": item_id})
                 if len(item_info) > 0:
+                    self.logger.debug('[insert_rating] Looking for the following info: %s', item_info)
                     for k, v in item.items():
                         if k in item_info and v is not None:  # sometimes the value IS None
+                            self.logger.debug("[insert_rating] Adding %sto info_used and create relative collections", k)
+
                             users_coll_name = self._coll_name(k, 'user')
                             items_coll_name = self._coll_name(k, 'item')
-                            self.info_used.add(k)
+
+                            self.db['utils'].update({"_id": 1},
+                                                    {"$addToSet": {'info_used': k}},
+                                                    upsert=True)
                             # see comments above
                             self.db['tot_' + users_coll_name].update({'_id': user_id},
                                                                      {'$inc': {v: float(rating)}},
@@ -372,85 +435,119 @@ class Recommender(object):
                 item_based = True
                 #Just take user_id for the user vector
                 df_user = pd.DataFrame(self.user_ratings).fillna(0).astype(int)[[user_id]]
-            if len(self.info_used) > 0:
-                for i in self.info_used:
+            info_used = self.info_used
+            if len(info_used) > 0:
+                for i in info_used:
                     if self.tot_categories_user_ratings[i].get(user_id):
                         info_based.append(i)
                         df_tot_cat_user[i] = pd.DataFrame(self.tot_categories_user_ratings[i]).fillna(0).astype(int)[[user_id]]
                         df_n_cat_user[i] = pd.DataFrame(self.n_categories_user_ratings[i]).fillna(0).astype(int)[[user_id]]
-        else:  # read if from Mongodb
+        else:  # Mongodb
+            # Did the user rate anything?
             if self.db['user_ratings'].find_one({"_id": user_id}):
                 item_based = True
-                #Take only user_id
-                df_user = pd.DataFrame.from_records(list(self.db['item_ratings'].find())).set_index('_id').fillna(0).astype(int)[[user_id]]
-            if len(self.info_used) > 0 and False: #XXX
-                for i in self.info_used:
+                try:
+                    df_user = pd.DataFrame.from_records(list(self.db['item_ratings'].find())).set_index('_id').fillna(0).astype(int)[[user_id]]
+                except:
+                    self.logger.warning("[get_recommendations. item and user ratings colls not synced")
+                    self._sync_user_item_ratings()
+                    df_user = pd.DataFrame.from_records(list(self.db['item_ratings'].find())).set_index('_id').fillna(0).astype(int)[[user_id]]
+            info_used = self.db['utils'].find_one({"_id": 1}, {'info_used': 1, "_id": 0}).get('info_used', [])
+            self.logger.debug("[get_recommendations] info_used: %s", info_used)
+            if len(info_used) > 0:
+                for i in info_used:
                     item_coll_name = self._coll_name(i, 'item')
-                    if self.db['tot_' + item_coll_name].find_one({"_id": user_id}):
+                    user_coll_name = self._coll_name(i, 'user')
+                    if self.db['tot_' + user_coll_name].find_one({"_id": user_id}):
                         info_based.append(i)
                         df_tot_cat_user[i] = pd.DataFrame.from_records(list(self.db['tot_' + item_coll_name].find())).set_index('_id').fillna(0).astype(int)[[user_id]]
                         df_n_cat_user[i] = pd.DataFrame.from_records(list(self.db['n_' + item_coll_name].find())).set_index('_id').fillna(0).astype(int)[[user_id]]
+                        self.logger.debug("[get_recommendations]. df_tot_cat_user[%s]:%s\n", i, df_tot_cat_user[i])
         if item_based:
-            if not fast or (time() - self.cooccurence_updated > 3600):
-                self._create_cooccurence()
             try:
                 # this might fail for fast in case a user has rated an item
                 # but the co-occurence matrix has not been updated
                 # therefore the matrix and the user-vector have different
                 # dimension
+                if not fast or (time() - self.cooccurence_updated > 3600):
+                    self._create_cooccurence()
                 rec = self._items_cooccurence.T.dot(df_user[user_id])
+                self.logger.debug("[get_recommendations] Rec: %s", rec)
             except:
-                self._create_cooccurence()
-                rec = self._items_cooccurence.T.dot(df_user[user_id])
+                self.logger.debug("[get_recommendations] 1st rec production failed, calling _create_cooccurence.")
+                try:
+                    self._create_cooccurence()
+                    rec = self._items_cooccurence.T.dot(df_user[user_id])
+                    self.logger.debug("[get_recommendations] Rec: %s", rec)
+                except:
+                    self.logger.warning("[get_recommendations] user_ and item_ratings seem not synced")
+                    self._sync_user_item_ratings()
+                    self._create_cooccurence()
+                    rec = self._items_cooccurence.T.dot(df_user[user_id])
+                    self.logger.debug("[get_recommendations] Rec: %s", rec)
+
             # Add to rec items according to popularity
             rec.sort(ascending=False)
 
             if len(rec) < max_recs:
-                self.compute_items_by_popularity(fast)
+                self.compute_items_by_popularity(fast=fast)
                 for v in self.items_by_popularity:
                     if len(rec) == max_recs:
                         break
                     elif v not in rec.index:
                         n = len(rec)
-                        rec.set_value(v, rec.values[n - 1]*n/(n+1))  # supposing score goes down according to Zipf distribution
+                        rec.set_value(v, rec.values[n - 1]*n/(n+1.))  # supposing score goes down according to Zipf distribution
         else:
-            self.compute_items_by_popularity(fast)
+            self.compute_items_by_popularity(fast=fast)
             for i, v in enumerate(self.items_by_popularity):
                 if len(rec) == max_recs:
                     break
-                rec.set_value(v, self.max_rating / (i+1))  # As comment above, starting from max_rating
+                rec.set_value(v, self.max_rating / (i+1.))  # As comment above, starting from max_rating
+        self.logger.debug("[get_recommendations] Rec after item_based or not: %s", rec)
 
         # Now, the worse case we have rec=popular with score starting from max_rating
         # and going down as 1/i (this is item_based == False)
 
         global_rec = rec.copy()
-        if len(self.info_used) > 0:
+        if len(info_used) > 0:
             cat_rec = {}
             for cat in info_based:
                 user_vec = df_tot_cat_user[cat][user_id] / df_n_cat_user[cat][user_id].replace(0, 1)
+                # print "DEBUG get_recommendations. user_vec:\n", user_vec
                 try:
                     cat_rec[cat] = self._categories_cooccurence[cat].T.dot(user_vec)
                     cat_rec[cat].sort(ascending=False)
+                    #self.logger.debug("[get_recommendations] cat_rec (try):\n %s", cat_rec)
                 except:
                     self._create_cooccurence()
                     cat_rec[cat] = self._categories_cooccurence[cat].T.dot(user_vec)
                     cat_rec[cat].sort(ascending=False)
+                    #self.logger.debug("[get_recommendations] cat_rec (except):\n %s", cat_rec)
                 for k, v in rec.iteritems():
+                    #self.logger.debug("[get_recommendations] rec_item_id: %s", k)
                     try:
                         if not self.db:
                             item_info_value = self.items[k][cat]
                         else:
-                            item_info_value = self.db['items'].find_one({"_id": k}, {"_id": 0, cat: 1})[cat]
-                        global_rec[k] = v + cat_rec[cat][item_info_value]
-                    except:
-                        print "ERROR on item", k, ", category", cat
+                            item_info_value = self.db['items'].find_one({"_id": k}, {"_id": 0, cat: 1}).get(cat)
+                        #self.logger.debug("DEBUG get_recommendations. item value for %s: %s", cat, item_info_value)
+                        # In case the info value is not in cat_rec (as it can obviously happen
+                        # because a rec'd item coming from most popular can have the value of
+                        # an info (author etc) which is not in the rec'd info
+                        if item_info_value:
+                            global_rec[k] = v + cat_rec.get(cat, []).get(item_info_value, 0)
+                    except Exception, e:
+                        self.logger.error("item %s, category %s", k, cat)
+                        logging.exception(e)
         global_rec.sort(ascending=False)
-        #return list(rec[df_user[user_id] == 0].index)
+        self.logger.debug("[get_recommendations] global_rec:\n %s", global_rec)
+
         if item_based:
-            # list(rec[df_user[user_id] == 0].index)[:max_recs],
-            return list(global_rec[df_user[user_id] == 0].index)[:max_recs]
+            # If the user has rated all items, return an empty list
+            rated = df_user[user_id] != 0
+            self.logger.debug("Rated: %s", rated)
+            return [i for i in global_rec.index if not rated.get(i, False)][:max_recs]
         else:
-            # list(rec.index)[:max_recs]
             return list(global_rec.index)[:max_recs]
 
 
